{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}

module J where

import Data.List
import Data.Maybe
import Parsing

class JShow a where
  jShow :: a -> String

instance JShow Int where
  jShow = show

instance {-# OVERLAPS #-} JShow String where
  jShow s = "\"" ++ concatMap (\c -> if c == '"' then "\"" else pure c) s ++ "\""

instance JShow Bool where
  jShow True = "true"
  jShow False = "false"

instance JShow a => JShow [a] where
  jShow xs = "[" ++ intercalate "," (map jShow xs) ++ "]"

instance JShow a => JShow (Maybe a) where
  jShow (Just x) = jShow x
  jShow Nothing = "null"

class JRead a where
  jRead :: Parser a

instance JRead Int where
  jRead = read <$> some digit <* ws

instance {-# OVERLAPS #-} JRead String where
  jRead = char '"' *> many (noneOf ['"', '\\'] <|> '\"' <$ string "\\\"" <|> '\\' <$ string "\\\\" <|> '\n' <$ string "\\\n") <* char '"' <* ws

instance JRead Bool where
  jRead = (True <$ string "true" <|> False <$ string "false") <* ws

instance JRead a => JRead [a] where
  jRead = char '[' *> ws *> (((:) <$> jRead <*> many (char ',' >> ws >> jRead)) <|> return []) <* char ']' <* ws

instance JRead a => JRead (Maybe a) where
  jRead = Nothing <$ string "null" <* ws <|> Just <$> jRead

jReadFull :: JRead a => String -> a
jReadFull = snd . fromJust . runParser (ws *> jRead <* eof)
