const std = @import("std");
const glue = @import("glue.zig");

const cell_size = 3;
const segment_size = 5;
const Cell = [cell_size][cell_size]bool;
const Board = [cell_size*segment_size*3+1][cell_size*segment_size*2+1]bool;
const Grid = [segment_size][segment_size]bool;

fn fromText(text: []const u8) Cell {
    var x: usize = 0;
    var y: usize = 0;
    var out: Cell = undefined;
    for (text) |c| {
        if (x == cell_size) {
            std.debug.assert(c == '\n');
            x = 0;
            y += 1;
            continue;
        }
        out[y][x] = switch (c) {
            '#' => true,
            ' ' => false,
            else => unreachable,
        };
        x += 1;
    }
    std.debug.assert(x == cell_size and y == cell_size-1);
    return out;
}

// see missingno.png
const white = fromText(
    \\   
    \\   
    \\   
);

const bad = fromText(
    \\###
    \\###
    \\###
);

const good_side = fromText(
    \\###
    \\   
    \\###
);

const good_down = fromText(
    \\# #
    \\# #
    \\# #
);

const good_bsls = fromText(
    \\ ##
    \\# #
    \\## 
);

const good_fsls = fromText(
    \\## 
    \\# #
    \\ ##
);

fn drawCell(board: *Board, cell: Cell, x: usize, y: usize) void {
    for (cell) |row, j| {
        for (row) |v, i| {
            board[y*cell_size+j][x*cell_size+i] = v;
        }
    }
}

fn drawSegment(board: *Board, grid: Grid, x: usize, y: usize, good: Cell) void {
    for (grid) |row, j| {
        for (row) |v, i| {
            drawCell(board, if (v) good else bad, x*segment_size+i, y*segment_size+j);
        }
    }
}

fn whiteSegment(board: *Board, x: usize, y: usize) void {
    var j: usize = 0;
    while (j < segment_size) : (j += 1) {
        var i: usize = 0;
        while (i < segment_size) : (i += 1) {
            drawCell(board, white, x*segment_size+i, y*segment_size+j);
        }
    }
}

fn rightSegment(board: *Board, x: usize, y: usize) void {
    const i = segment_size-1;
    var j: usize = 0;
    while (j < segment_size) : (j += 1) {
        drawCell(board, bad, x*segment_size+i, y*segment_size+j);
    }
}

fn write(buf: *[]u8, s: []const u8) void {
    std.mem.copy(u8, buf.*, s);
    buf.* = buf.*[s.len..];
}

fn isLine(flat_grid: [25]bool) bool {
    const grid = @bitCast(Grid, flat_grid);

    var board = std.mem.zeroes(Board);
    rightSegment(&board, 0, 0);
    whiteSegment(&board, 0, 1);
    drawSegment(&board, grid, 1, 0, good_fsls);
    drawSegment(&board, grid, 1, 1, good_side);
    drawSegment(&board, grid, 1, 2, good_bsls);
    drawSegment(&board, grid, 0, 2, good_down);

    var buf_back: [8192]u8 = undefined;
    var buf: []u8 = buf_back[0..];
    write(&buf, "canSolve [");
    for (board) |row| {
        write(&buf, "[");
        for (row) |e| {
            write(&buf, if (e) "True,"[0..] else "False,"[0..]);
        }
        // lol
        (buf.ptr-1).* = ']';
        write(&buf, ",");
    }
    (buf.ptr-1).* = ']';

    var arena = std.heap.ArenaAllocator.init(std.heap.page_allocator);
    defer arena.deinit();

    return glue.glueCall(bool, "CanSolve.hs", buf_back[0..@ptrToInt(buf.ptr)-@ptrToInt(&buf_back)], arena.allocator());
}
