const std = @import("std");
const zlib = @cImport({
   @cInclude("zlib.h"); 
});

pub fn compress(data: []const u8) ![]u8 {
    var s: zlib.z_stream = undefined;
    s.zalloc = null;
    s.zfree = null;
    s.@"opaque" = null;
    s.avail_in = @intCast(c_uint, data.len);
    s.next_in = @intToPtr(?*u8, @ptrToInt(data.ptr));
    if (zlib.deflateInit(&s, zlib.Z_DEFAULT_COMPRESSION) != zlib.Z_OK) {
        return error.InitFailed;
    }
    defer _ = zlib.deflateEnd(&s);
    var len_buf = try std.heap.c_allocator.alloc(u8, data.len+128);
    std.mem.writeIntLittle(u64, len_buf[0..8], @intCast(u64, data.len));
    var buf = len_buf[8..];
    s.avail_out = @intCast(c_uint, buf.len);
    s.next_out = @intToPtr(?*u8, @ptrToInt(buf.ptr));
    if (zlib.deflate(&s, zlib.Z_FINISH) != zlib.Z_STREAM_END) {
        return error.WtfZlib;
    }
    len_buf = std.heap.c_allocator.realloc(len_buf, len_buf.len-s.avail_out) catch buf;
    return len_buf;
}

pub fn decompress(len_data: []const u8) ![]u8 {
    const len = @intCast(usize, std.mem.readIntLittle(u64, len_data[0..8]));
    const data = len_data[8..];
    var s: zlib.z_stream = undefined;
    s.zalloc = null;
    s.zfree = null;
    s.@"opaque" = null;
    s.avail_in = @intCast(c_uint, data.len);
    s.next_in = @intToPtr(*u8, @ptrToInt(data.ptr));
    if (zlib.inflateInit(&s) != zlib.Z_OK) {
        return error.InitFailed;
    }
    defer _ = zlib.inflateEnd(&s);
    var buf = try std.heap.c_allocator.alloc(u8, len);
    s.avail_out = @intCast(c_uint, buf.len);
    s.next_out = @intToPtr(*u8, @ptrToInt(buf.ptr));
    if (zlib.inflate(&s, zlib.Z_FINISH) != zlib.Z_STREAM_END) {
        return error.NiceZlib;
    }
    buf = std.heap.c_allocator.realloc(buf, buf.len-s.avail_out) catch buf;
    return buf;
}
