import (
	"strings"
	"os"
	"unsafe"
	"encoding/gob"
	"fmt"
	"io"
)

const (
	UP = 1
	RIGHT = 2
	DOWN = 3
	LEFT = 4
)

// using []uint8 for the path makes Go assume it's a bytestring and marshal it as Base64 (???), so we use uint16
type dir uint16

func contains(unsorted []int, x int) bool {
	if len(unsorted) == 0 {
		return false
	}
	s := GlueCall[[]int]("r1.hs", strings.ReplaceAll(fmt.Sprintf("entry%v", unsorted), " ", ","))
	for len(s) > 1 {
		mid := len(s) / 2
		if x < s[mid] {
			s = s[:mid]
		} else {
			s = s[mid:]
		}
	}
	return s[0] == x
}

type pos struct {
	X int
	Y int
}

type state struct {
	Maze []bool
	Width int
	Height int
	Frontier []pos
	Crumbs []dir
	Seen []int
}

func load(maze []bool, width int, height int) state {
	filename := Entry(*(*[]byte)(unsafe.Pointer(&maze)))
	file, err := os.Open(filename)
	if err != nil {
		return state{maze, width, height, []pos{pos{0, 0}}, make([]dir, len(maze)), []int{}}
	}
	dec := gob.NewDecoder(file)
	var s state
	dec.Decode(&s)
	return s
}

func save(s *state) {
	filename := Entry(*(*[]byte)(unsafe.Pointer(&s.Maze)))
	file, _ := os.Create(filename)
	defer file.Close()
	enc := gob.NewEncoder(file)
	enc.Encode(*s)

    // progress indicator
	var clear float64 = 0
	for _, c := range s.Maze {
		if !c {
			clear++
		}
	}
	progress := float64(len(s.Seen)) / clear

	prog, _ := os.OpenFile("progress", os.O_WRONLY | os.O_CREATE, 0644)
	defer prog.Close()
	prog.Truncate(6)
	prog.Seek(6, io.SeekStart)
	fmt.Fprintf(prog, "current %.2f%%", progress*100.0)
}

func move(p pos, d dir, v int) pos {
	switch d {
	case UP:
		p.Y -= v
	case RIGHT:
		p.X += v
	case DOWN:
		p.Y += v
	case LEFT:
		p.X -= v
	}
	return p
}

func try(s *state, d dir) {
    p := move(s.Frontier[0], d, 1)
    x := p.X
    y := p.Y
    index := y*s.Width+x
	if x >= 0 && x < s.Width && y >= 0 && y < s.Height && !s.Maze[index] {
		if s.Crumbs[index] == 0 {
			s.Frontier = append(s.Frontier, pos{x, y})
			s.Crumbs[index] = d;
		}
	}
}

func entry(maze []bool, width int) []dir {
	if len(maze) % width != 0 {
		panic("?")
	}
	height := len(maze) / width

    s := load(maze, width, height)

	for ; len(s.Frontier) > 0; s.Frontier = s.Frontier[1:] {
		p := s.Frontier[0]
		index := p.Y*width+p.X
		if contains(s.Seen, index) {
			continue
		}
		s.Seen = append(s.Seen, index)

		if p == (pos{width-1, height-1}) {
			seq := []dir{}
			for p != (pos{0, 0}) {
				d := s.Crumbs[p.Y*width+p.X]
				seq = append(seq, d)
				p = move(p, d, -1)
			}
			// lol what
			for i, j := 0, len(seq)-1; i < j; i, j = i+1, j-1 {
				seq[i], seq[j] = seq[j], seq[i]
			}
			return seq
		}

		try(&s, UP)
		try(&s, RIGHT)
		try(&s, DOWN)
		try(&s, LEFT)

		save(&s)
	}

	return nil
}
